**************************************************************************;
** Macro name      : ncspline                                           **;
** Version number  : 1.02                                               **;
** Authors         : Alberto Garcia-Hernandez                           **;
**                   Dimitris Rizopoulos                                **;
** Date            : 23-Jan-2015                                        **;
** SAS version     : 9.0 or higher                                      **;
** Purpose         : Spline Basis functions for Natural Cubic           **;
**                   splines                                            **;   
** Input Data      : A dataset with a variable T                        **;
** Input Parameters: indata       = Input Dataset                       **;
**                   outdata      = Output Dataset                      **;
**                   invariable   = Variable that stores T              **;
**                   outarray     = Array/Matrix to store the basis     **;
**                   dimout = outarray  is a matrix of dimension        **;
**                            Dimeout x Number of basis functions       **; 
**                            use dimout=1 for the basis of variable T  **;
**                            use dimout>1 when bases are required for a**;
**                            vector of times                           **;  
**                   posout = position of the basis in the output matrix**;
**                   lowerknot= lower boundary knot in hex16. format    **;
**                   upperknot = upper boundary knot  in hex16. format  **;
**                   Nknots    = number of internal knots (max=3)       **;
**                   knot1, knot2, knot3 = internal knots value         **;
**                               in hex16. format     				    **;
**                   incderiv = YES if derivatives of the functions are **;
**                             required                                 **;
**                   incderiv2= YES if 2nd derivatives of the functions **;
**                             required                                 **;
**                   dif      = difference used to derivatives          **;
** Output          : Variable/Matrix with the Bspline Basis functions   **;
**                                                                      **;
** Further Details : See JM Macro Manuscript at JSS                     **;
**                        MACRO HISTORY                                 **;
** Version     Date        Changes                                      **;
** 1.01        15/12/2014  First Production Version                     **;
** 1.02        23/02/2015  New checks added for input knots not in      **;
**                         16-characters Hex16. format.                 **;
**                                                                      **;
** Copyright (C) 2014      Alberto Garcia Hernandez                     **;
**                         Dimitris Rizopoulos                          **;
** This program is free software: you can redistribute it and/or modify **;
** it under the terms of the GNU General Public License as published by **;
** the Free Software Foundation, either version 3 of the License, or    **;
** (at your option) any later version.                                  **;
**                                                                      **;
** This program is distributed in the hope that it will be useful,      **;
** but WITHOUT ANY WARRANTY, without even the implied warranty of       **;
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        **;
** GNU General Public License for more details.                         **;
**                                                                      **;
** You should have received a copy of the GNU General Public License    **;
** along with this program.  If not, see <http://www.gnu.org/licenses/  **;
** Everyone is permitted to copy and distribute verbatim copies         **;
** of this program, but changing it is not allowed.                     **;
**                                                                      **;
** Contact email: alberto.garciahernandez@astellas.com                  **;
**************************************************************************;

%macro ncspline(indata=, outdata=, invariable=, 
               lowerknot=, upperknot=, outarray=, dimout=1, posout=1, Nknots = 0,
               knot1=, knot2=, knot3=, knot4=, knot5=,knot6=, knot7=, knot8=, knot9=, 
               incderiv=NO, incderiv2=NO, dif=0.001, nparam_var=longi_nparam, if=);

%let nparam=%eval(2 + &Nknots.);     *number of parameters;

%if &nparam_var. NE %then %do;
  %let &nparam_var. = &nparam.;
%end;

%local datamin datamax;

%if %length(&lowerknot.) NE 0 and %length(&lowerknot.) NE 16 %then %do;
   %PUT %str(ER)ROR: Lowernot should be given in a 16-characters Hex16. format. ; 
   %ABORT;  
%end;
%if %length(&upperknot.) NE 0 and %length(&upperknot.) NE 16 %then %do;
   %PUT %str(ER)ROR: Upperknot should be given in a 16-characters Hex16. format.; 
   %ABORT;  
%end;
%if &Nknots. >0 %then %do;
  %do m = 1 %to &Nknots.;
     %if %length(&&knot&m.) NE 0 and %length(&&knot&m.) NE 16 %then %do;
        %PUT %str(ER)ROR: Knot &m. should be given in a 16-characters Hex16. format.; 
        %ABORT;  
     %end;
  %end;
%end;

 %if %upcase(&indata.) NE _INSIDE %then %do;

  proc sql noprint;
   select put(min(&invariable.),hex16.), put(max(&invariable.),hex16.) into :datamin, :datamax
   from &indata.;
 quit;
 %if &lowerknot. = %then %do; %let lowerknot = &datamin.; %end;
 %if &upperknot. = %then %do; %let upperknot = &datamax.; %end;

%put Boundary knots are:  %sysfunc(inputn(&lowerknot.,hex16.)) (&lowerknot.) and %sysfunc(inputn(&upperknot.,hex16.)) (&upperknot.);


  data &outdata.;
    set &indata.;
 %end;

 array &outarray.[&dimout.,%eval(&nparam.)] &outarray.0-&outarray.%eval(&dimout.*&nparam.-1); 
  %if &if. NE %then if &if then do;;
 if &invariable. ne . then &outarray.[&posout.,1] = 1;
 &outarray.[&posout.,2] = &invariable.; 

 %if &Nknots. = 0 %then %do;
     %PUT %str(ER)ROR: THE NATURAL CUBIC SPLINES FUNCTION REQUIRE AT LEAST 1 KNOT IN ADDITION TO THE BOUNDARY KNOTS;%ABORT;
   %end;
   %else %if &Nknots. >0 %then %do;
     array d[%eval(&Nknots.+1)]; 
      d[1] =                             (  (((&invariable.)>input("&lowerknot.",hex16.)) * (&invariable.-input("&lowerknot.",hex16.))**3) 
	                                       -(((&invariable.)>input("&upperknot.",hex16.)) * (&invariable.-input("&upperknot.",hex16.))**3) 
                                           )/ (input("&upperknot.",hex16.) - input("&lowerknot.",hex16.));
     %do k = 1 %to &Nknots.;
    
      d[%eval(&k+1)] =                         (  (((&invariable.)>input("&&knot&k.",hex16.)) * (&invariable.-input("&&knot&k.",hex16.))**3) 
	                                       -(((&invariable.)>input("&upperknot.",hex16.)) * (&invariable.-input("&upperknot.",hex16.))**3) 
                                           )/ (input("&upperknot.",hex16.)- input("&&knot&k.",hex16.));
      %end;

	 %put Lower Knot  is :  %sysfunc(inputn(&lowerknot.,hex16.)) (&lowerknot.);
	 %put Upper Knot  is :  %sysfunc(inputn(&upperknot.,hex16.)) (&upperknot.);
	  %do k=1 %to &Nknots.;
	   %put Knot &k. is :  %sysfunc(inputn(&&knot&k.,hex16.)) (&&knot&k.);
         &outarray.[&posout.,%eval(&k.+2)] = d[%eval(&k.)] - d[%eval(&Nknots.+1)];
	  %end;	   

   %end;
   %if &if. NE %then end;;
   drop d1--d%eval(&Nknots.+1);

 %if %upcase(&indata.) NE _INSIDE %then %do;
  run;
 %end;

%if &incderiv. = YES and %upcase(&indata.) NE _INSIDE %then %do;

%ncspline(indata=&outdata., outdata=&outdata., invariable=&invariable.-&dif., outarray=delta_&outarray.,  posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., if=&if., nparam_var=&nparam_var.);

%ncspline(indata=&outdata., outdata=&outdata., invariable=&invariable.+&dif., outarray=delta2_&outarray.,  posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., if=&if., nparam_var=&nparam_var.);

%if &incderiv2 = YES %then %do;
%ncspline(indata=&outdata., outdata=&outdata., invariable=&invariable.-2*&dif., outarray=tdelta_&outarray.,  posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., if=&if., nparam_var=&nparam_var.);

%ncspline(indata=&outdata., outdata=&outdata., invariable=&invariable.+2*&dif., outarray=tdelta2_&outarray.,  posout=&posout., dimout=&dimout.,
            lowerknot=&lowerknot., upperknot=&upperknot., Nknots=&Nknots., knot1=&knot1., knot2=&knot2., knot3=&knot3.,
            knot4=&knot4.,knot5=&knot5.,knot6=&knot6.,knot7=&knot7.,knot8=&knot8.,knot9=&knot9., if=&if., nparam_var=&nparam_var.);
%end;

data &outdata.;
  set &outdata.;
   array &outarray.[&dimout.,%eval(&nparam.)]  &outarray.0-&outarray.%eval(&dimout.*&nparam.-1); 
   array delta_&outarray.[&dimout.,%eval(&nparam.)]    delta_&outarray.0 - delta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array delta2_&outarray.[&dimout.,%eval(&nparam.)]  delta2_&outarray.0 -  delta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array drv_&outarray.[&dimout.,%eval(&nparam.)] drv_&outarray.0 - drv_&outarray.%eval(&dimout.*&nparam.-1) ; 

%if &incderiv2 = YES %then %do;
   array tdelta_&outarray.[&dimout.,%eval(&nparam.)]    tdelta_&outarray.0 -   tdelta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array tdelta2_&outarray.[&dimout.,%eval(&nparam.)]   tdelta2_&outarray.0 -  tdelta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array secdrv_&outarray.[&dimout.,%eval(&nparam.)] secdrv_&outarray.0 - secdrv_&outarray.%eval(&dimout.*&nparam.-1) ; 
%end;

data &outdata.;
  set &outdata.;
   array &outarray.[&dimout.,%eval(&nparam.)]  &outarray.0-&outarray.%eval(&dimout.*&nparam.-1); 
   array delta_&outarray.[&dimout.,%eval(&nparam.)]    delta_&outarray.0 - delta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array delta2_&outarray.[&dimout.,%eval(&nparam.)]  delta2_&outarray.0 -  delta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array drv_&outarray.[&dimout.,%eval(&nparam.)] drv_&outarray.0 - drv_&outarray.%eval(&dimout.*&nparam.-1) ; 

%if &incderiv2 = YES %then %do;
   array tdelta_&outarray.[&dimout.,%eval(&nparam.)]    tdelta_&outarray.0 -   tdelta_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array tdelta2_&outarray.[&dimout.,%eval(&nparam.)]   tdelta2_&outarray.0 -  tdelta2_&outarray.%eval(&dimout.*&nparam.-1) ; 
   array secdrv_&outarray.[&dimout.,%eval(&nparam.)] secdrv_&outarray.0 - secdrv_&outarray.%eval(&dimout.*&nparam.-1) ; 
%end;

  if &invariable.>input("&datamin.",hex16.) and &invariable.<input("&datamax.",hex16.) then do;
     do m = 1 to &nparam.;
	   drv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m] - delta_&outarray.[&posout.,m])/(2*&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m]   - 2*&outarray.[&posout.,m]  + delta_&outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;
    else if &invariable.=input("&datamin.",hex16.) then do;
     do m = 1 to &nparam.;
	  drv_&outarray.[&posout.,m]=(delta2_&outarray.[&posout.,m] - &outarray.[&posout.,m])/(&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(tdelta2_&outarray.[&posout.,m] -2*delta2_&outarray.[&posout.,m] + &outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;
	 else if &invariable.=input("&datamax.",hex16.) then do;
     do m = 1 to &nparam.;
      drv_&outarray.[&posout.,m]=(&outarray.[&posout.,m]-delta_&outarray.[&posout.,m] )/(&dif.);   
	  %if &incderiv2 = YES %then secdrv_&outarray.[&posout.,m]=(&outarray.[&posout.,m] - 2*delta_&outarray.[&posout.,m] + tdelta_&outarray.[&posout.,m])/(&dif.**2);;   
     end;  
    end;

 drop  m delta_&outarray.0 - delta_&outarray.%eval(&dimout.*&nparam.-1)  delta2_&outarray.0 -  delta2_&outarray.%eval(&dimout.*&nparam.-1);
 %if &incderiv2 = YES %then drop  tdelta_&outarray.0 - tdelta_&outarray.%eval(&dimout.*&nparam.-1)  tdelta2_&outarray.0 -  tdelta2_&outarray.%eval(&dimout.*&nparam.-1);;

run;
%end;


%mend ncspline;
